//#define DBG
/*====================================================================*
 * @(#)cami2c.c
 *====================================================================*
 * camra_i2c
 *
 * https://raw.githubusercontent.com/raspberrypi/raspiraw/master/camera_i2c
 *====================================================================*
 * rpi3-gpiovirtbuf.c -- Control Raspberry Pi 3's activity LED by using the Mailbox interface
 *
 * Copyright (c) 2016 Sugizaki Yukimasa
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * https://raw.githubusercontent.com/6by9/rpi3-gpiovirtbuf/master/rpi3-gpiovirtbuf.c
 *====================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/ioctl.h>
//#include "raspberrypi-firmware.h"

static const char __PROGRAM_VERSION__[] = "@(#)camera_i2c.c Ver. 0.0009 (2022/01/31 10:21:19)";

#define RPI_FIRMWARE_DEV "/dev/vcio"
#define IOCTL_RPI_FIRMWARE_PROPERTY _IOWR(100, 0, char*)

#define REVISION_FILE "/proc/cpuinfo"
#define KERNEL_FILE   "/proc/sys/kernel/osrelease"

//---------------------------------------------------------------------
#define RPI_FIRMWARE_STATUS_REQUEST  (0) 
#define RPI_FIRMWARE_STATUS_SUCCESS  (0x80000000)
#define RPI_FIRMWARE_STATUS_ERROR    (0x80000001)
#define RPI_FIRMWARE_PROPERTY_END    (0)
#define RPI_FIRMWARE_SET_GPIO_STATE  (0x00038041)
//---------------------------------------------------------------------

#define revcmp(s,t) ((strcmp((s),(t))==0)?1:0)

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
char *
skipspace(char *s)
{
    while (isspace(*s)) {
        s++;
    }
    return  s;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
void
striptail(char *s)
{
    char    *p;

    for (p = s + strlen(s) - 1; isspace(*p); p--) {
        *p = '\0';
    }
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
char *
GetRev()
{
    FILE    *fp;
    char    buf[BUFSIZ], *p;
static char revision[16];

    if ((fp = fopen(REVISION_FILE, "r")) == NULL) {
        fprintf(stderr, "cami2c: can't open %s.\n", REVISION_FILE);
        exit(1);
    }
    revision[0] = '\0';
    while (fgets(buf, BUFSIZ, fp) != NULL) {
        if ((p = strstr(buf, "Revision")) != NULL) {
            striptail(p);
            p = strchr(p, ':');
            p = skipspace(p+1);
            strcpy(revision, p);
            break;
        }
    }
    fclose(fp);

    return  (revision[0] == 0) ? NULL : revision;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
char *
GetKernel()
{
    FILE    *fp;
    char    buf[BUFSIZ];
static char kernel[32];

    if ((fp = fopen(KERNEL_FILE, "r")) == NULL) {
        fprintf(stderr, "cami2c: can't open %s.\n", KERNEL_FILE);
        exit(1);
    }
    kernel[0] = '\0';
    if (fgets(buf, BUFSIZ, fp) != NULL) {
        striptail(buf);
        strcpy(kernel, buf);
    }
    fclose(fp);

    return  (kernel[0] == 0) ? NULL : kernel;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
static void
gpio_set(int gpio, int state)
{
    uint32_t buf[8];
    int fd;

    fd = open(RPI_FIRMWARE_DEV, O_NONBLOCK);
    if (fd == -1) {
        fprintf(stderr, "gpio_set: can't open %s\n", RPI_FIRMWARE_DEV);
        exit(1);
    }

    buf[0] = sizeof(buf);
    buf[1] = RPI_FIRMWARE_STATUS_REQUEST;
    buf[2] = RPI_FIRMWARE_SET_GPIO_STATE;
    buf[3] = 8;                            //data(gpio+state) size;
    buf[4] = 0;                            // req_resp_size
    buf[5] = gpio;                         // req_resp_size
    buf[6] = (state) ? 1 : 0;              // req_resp_size
    buf[7] = RPI_FIRMWARE_PROPERTY_END;

    if (ioctl(fd, IOCTL_RPI_FIRMWARE_PROPERTY, buf) < 0) {
        fprintf(stderr, "gpio_set: ioctl error\n");
        exit(1);
    }
    if (buf[1] != RPI_FIRMWARE_STATUS_SUCCESS) {
        fprintf(stderr, "gpio_set: RPi firmware returned 0x%08x\n", buf[1]);
        exit(1);
    }

    close(fd);
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
void
cm4camera_i2c(int mode)
{
    system("raspi-gpio set 0 ip");
    system("raspi-gpio set 1 ip");
    system("raspi-gpio set 44 ip");
    system("raspi-gpio set 44 a1");
    system("raspi-gpio set 45 ip");
    system("raspi-gpio set 45 a1");
    gpio_set(133, 1);
    system("i2cdetect -y 0 > /dev/null");

    if (mode == 1) {
        system("i2cdetect -y 10 > /dev/null");
    }
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
static void
cam_i2c()
{
    char *rev;
    char *ker; 
    char board_type[3];

    if ((rev = GetRev()) == NULL) {
        fprintf(stderr, "Revisoion not found.\n");
        exit(1);
    }
//    printf("rev=%s\n", rev);

    if ((ker = GetKernel()) == NULL) {
        fprintf(stderr, "Kernel version not found.\n");
        exit(1);
    }
//    printf("ker=%s\n", ker);

    board_type[0] = rev[3];
    board_type[1] = rev[4];
    board_type[2] = '\0';

#ifdef DBG
printf("rev       :%s\n", rev);
printf("kernel    :%s\n", ker);
printf("board_type:%s\n", board_type);
#endif

    if (revcmp(rev, "0002")
     || revcmp(rev, "0003")) { //case '0002'|'0003')
#ifdef DBG
        printf("B Rev1 - I2C 1 on GPIOs 2 & 3. GPIOs 5 & 27 for LED and power\n");
#endif
        // i2c on these pins
        system("raspi-gpio set 0 ip");
        system("raspi-gpio set 0 a0");
        system("raspi-gpio set 1 ip");
        system("raspi-gpio set 1 a0");
        //shutdown
        system("raspi-gpio set 27 op dh");
        //LED
        system("raspi-gpio set 5 op dh");
        //
#ifdef DBG
        printf("Use 'raspiraw -y 1' to communicate with the sensor\n");
#endif

    } else if (revcmp(rev, "0004")
            || revcmp(rev, "0005")
            || revcmp(rev, "0006")
            || revcmp(rev, "000d")
            || revcmp(rev, "000e")
            || revcmp(rev, "000f")) { //'0004'|'0005'|'0006'|'000d'|'000e'|'000f')
#ifdef DBG
        printf("B Rev2 - I2C 0 on GPIOs 0 & 1. GPIOs 5 & 21 for LED and power\n");
#endif
        system("raspi-gpio set 0 ip");
        system("raspi-gpio set 0 a0");
        system("raspi-gpio set 1 ip");
        system("raspi-gpio set 1 a0");
        //shutdown
        system("raspi-gpio set 21 dh");
        //LED
        system("raspi-gpio set 5 dh");
    } else if (revcmp(rev, "0010")
            || revcmp(rev, "0012")
            || revcmp(rev, "0013")
            || revcmp(rev, "1041")) { //'0010'|'0012'|'0013'|'1041')
#ifdef DBG
        printf("A+, B+, B2, B2 (2837), and Zero (not W) all revisions");
        printf(" - I2C 0 on GPIOs 28 & 29. GPIOs 32 & 41 for LED and power\n");
#endif
        // i2c can be on pins 0 and 1, so make sure they are not set to alt0
        system("raspi-gpio set 0 ip");
        system("raspi-gpio set 1 ip");
        // i2c on these pins
        system("raspi-gpio set 28 ip");
        system("raspi-gpio set 28 a0");
        system("raspi-gpio set 29 ip");
        system("raspi-gpio set 29 a0");
        //shutdown
        system("raspi-gpio set 41 dh");
        //LED
        system("raspi-gpio set 32 dh");

    } else if (revcmp(rev, "0011")
            || revcmp(rev, "0014")) { //'0011'|'0014')
#ifdef DBG
        printf("Compute Module/ CM3 - I2C 0 on GPIO 0 & 1. \n");
        printf("GPIOs 2 & 3 for LED and power\n");
        printf("Warning: This is just a default setting for cam1.\n");
        printf("Download dt-blob.bin from https://www.raspberrypi.org/");
        printf("documentation/hardware/computemodule/dt-blob-cam1.bin ");
        printf("and place it in /boot/.\n");
        printf("For cable connection, see https://www.raspberrypi.org/");
        printf("documentation/hardware/computemodule/cmio-camera.md\n");
#endif
        // i2c can be on pins 28 and 29, so make sure they are not set to alt0
        system("raspi-gpio set 28 ip");
        system("raspi-gpio set 29 ip");
        // i2c on these pins
        system("raspi-gpio set 0 ip");
        system("raspi-gpio set 0 a0");
        system("raspi-gpio set 1 ip");
        system("raspi-gpio set 1 a0");
        // shutdown
        system("raspi-gpio set 3 out");
        system("raspi-gpio set 3 dh");
        // LED
        system("raspi-gpio set 2 out");
        system("raspi-gpio set 2 dh");
    } else {
        // New format revision encodings
        if (revcmp(board_type, "02")
         || revcmp(board_type, "03")
         || revcmp(board_type, "04")
         || revcmp(board_type, "09")) {// '02'|'03'|'04'|'09')
#ifdef DBG
            printf("A+, B+, B2, B2 (2837), and Zero (not W) all revisions ");
            printf("- I2C 0 on GPIOs 28 & 29. GPIOs 32 & 41 for LED and power\n");
#endif
            // i2c can be on pins 0 and 1, so make sure they are not set to alt0
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 1 ip");
            // i2c on these pins
            system("raspi-gpio set 28 ip");
            system("raspi-gpio set 28 a0");
            system("raspi-gpio set 29 ip");
            system("raspi-gpio set 29 a0");
            //shutdown
            system("raspi-gpio set 41 dh");
            //LED
            system("raspi-gpio set 32 dh");
        } else if (revcmp(board_type, "0c")) {// '0c')
#ifdef DBG
            printf("PiZero W - I2C 0 on GPIO 28 & 29. GPIOs 40 & 44 ");
            printf("for LED and power\n");
#endif
            // i2c can be on pins 0 and 1, so make sure they are not set to alt0
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 1 ip");
            // i2c on these pins
            system("raspi-gpio set 28 ip");
            system("raspi-gpio set 28 a0");
            system("raspi-gpio set 29 ip");
            system("raspi-gpio set 29 a0");
            // shutdown
            system("raspi-gpio set 44 dh");
            // LED
            system("raspi-gpio set 40 dh");
        } else if (revcmp(board_type, "0a")) {// '0a')
#ifdef DBG
            printf("Compute Module/ CM3 - I2C 0 on GPIO 0 & 1. ");
            printf("GPIOs 2 & 3 for LED and power\n");
            printf("Warning: This is just a default setting for cam1.\n");
            printf("Download dt-blob.bin from https://www.raspberrypi.org/");
            printf("documentation/hardware/computemodule/dt-blob-cam1.bin and place it in /boot/.\n");
            printf("For cable connection, see https://www.raspberrypi.org/");
            printf("documentation/hardware/computemodule/cmio-camera.md\n");
#endif
            // i2c can be on pins 28 and 29, so make sure they are not set to alt0
            system("raspi-gpio set 28 ip");
            system("raspi-gpio set 29 ip");
            // i2c on these pins
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 0 a0");
            system("raspi-gpio set 1 ip");
            system("raspi-gpio set 1 a0");
            // shutdown
            system("raspi-gpio set 3 out");
            system("raspi-gpio set 3 dh");
            // LED
            system("raspi-gpio set 2 out");
            system("raspi-gpio set 2 dh");
        } else if (revcmp(board_type, "08")
                || revcmp(board_type, "0d")
                || revcmp(board_type, "0e")) {// '08'|'0d'|'0e')
#ifdef DBG
            printf("Raspberry Pi3B / Pi3B+ / 3A\n");
#endif
            // https://www.raspberrypi.org/forums/viewtopic.php?f=38&t=120702&start=100
            // Pins 44&45 Alt1=i2c0, alt2=i2c1
            // i2c0 can be on pins 0 and 1, so make sure they are not set to alt0
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 1 ip");
            system("raspi-gpio set 28 ip");
    	    system("raspi-gpio set 29 ip");
            // i2c on these pins
            system("raspi-gpio set 44 ip");
            system("raspi-gpio set 44 a1");
            system("raspi-gpio set 45 ip");
            system("raspi-gpio set 45 a1");
            //shutdown
            // you need this one: https://github.com/6by9/rpi3-gpiovirtbuf
            //system("raspi-gpio set 41 1");
            gpio_set(133, 1); // `dirname $0`/rpi3-gpiovirtbuf s 133 1
            //LED
            //system("raspi-gpio set 32 1");
        } else if (revcmp(board_type, "11")) {// '11')
#ifdef DBG
            printf("4B(1G/2G/4G/8G)\n");
#endif
            // https://www.raspberrypi.org/forums/viewtopic.php?f=38&t=120702&start=100
            // Pins 44&45 Alt1=i2c0, alt2=i2c1
            // i2c0 can be on pins 0 and 1, so make sure they are not set to alt0
            // Do NOT alter 28&29 as they are used by the ethernet PHY.
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 1 ip");
            // i2c on these pins
            system("raspi-gpio set 44 ip");
            system("raspi-gpio set 44 a1");
            system("raspi-gpio set 45 ip");
            system("raspi-gpio set 45 a1");
            //shutdown
            // you need this one: https://github.com/6by9/rpi3-gpiovirtbuf
            //system("raspi-gpio set 41 1");
            gpio_set(133, 1); //`dirname $0`/rpi3-gpiovirtbuf s 133 1
            //LED
            //system("raspi-gpio set 32 1");
        } else if (revcmp(board_type, "12")) {// '12')
#ifdef DBG
            printf("PiZero2W\n");
#endif
            // Pins 44&45 Alt1=i2c0, alt2=i2c1
            // i2c0 can be on pins 0 and 1, so make sure they are not set to alt0
            // Do NOT alter 28&29 as they are used by the ethernet PHY.
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 1 ip");
            // i2c on these pins
            system("raspi-gpio set 44 ip");
            system("raspi-gpio set 44 a1");
            system("raspi-gpio set 45 ip");
            system("raspi-gpio set 45 a1");
            //shutdown
            system("raspi-gpio set 40 dh");
            // No LED line.
        } else if (revcmp(board_type, "13")) {// '13')
            // Pi400
#ifdef DBG
            printf("Failed: Pi400 doesn't have a camera connector\n");
#endif
        } else if (revcmp(board_type, "14")) {// '14')
            // CM4
#ifdef DBG
            printf("CM4 - assume CAM1 on 44&45\n");
#endif
#if  0
            // i2c can be on pins 44 and 45, so make sure they are not set to alt0
            system("raspi-gpio set 44 ip");
            system("raspi-gpio set 45 ip");
            // i2c on these pins
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 0 a0");
            system("raspi-gpio set 1 ip");
            system("raspi-gpio set 1 a0");
            // shutdown
            gpio_set(133, 1); //`dirname $0`/rpi3-gpiovirtbuf s 133 1
#else
            system("raspi-gpio set 0 ip");
            system("raspi-gpio set 1 ip");
            system("raspi-gpio set 44 ip");
            system("raspi-gpio set 44 a1");
            system("raspi-gpio set 45 ip");
            system("raspi-gpio set 45 a1");
            gpio_set(133, 1);
#endif
        } else {
            fprintf(stderr, "Failed: don't know how to set GPIO for this board! ");
            fprintf(stderr, "Type is %s(%s)\n", rev, board_type);
        }
    }

//    if [ "$kernel" = "4" ]
//    then
//        printf("Use i2c-0 for the sensor (-y 0)\n");
//    else
//        printf("Use i2c-10 for the sensor (-y 10)\n");
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
void
camera_i2c(int device_no)
{
    cam_i2c();
    if (device_no == 0) {
        system("i2cdetect -y 0 > /dev/null");
        //system("i2cdetect -y 0");
    } else {
        system("i2cdetect -y 10 > /dev/null");
    }
}

#ifdef MAIN
/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
main(int ac, char *av[])
{
    camera_i2c(0);
    return  0;
}
#endif
